using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Storage;
using System.Diagnostics;

namespace UWB_XNALib
{
    public class RGraphicsDeviceService : IGraphicsDeviceService
    {
        private GraphicsDevice m_GraphicsDevice;

        public event EventHandler DeviceCreated;
        public event EventHandler DeviceDisposing;
        public event EventHandler DeviceReset;
        public event EventHandler DeviceResetting;

        public GraphicsDevice GraphicsDevice
        {
            get { return this.m_GraphicsDevice; }
            set { m_GraphicsDevice = value; }
        }
    }

    public class RGameView : IServiceProvider
    {
        public RGraphicsDeviceService m_graphicsDeviceService;

        public void SetService(RGraphicsDeviceService s)
        {
            m_graphicsDeviceService = s;
        }

        public object GetService(System.Type serviceType)
        {
            return this.m_graphicsDeviceService;
        }
    }
    //This is our custome vertex structure XNA likes to recieve 
    //arrays of structs with vertex information. For simple, unlit
    //drawing; a vertex and color at the vertex will suffice.
    public struct DeviceVertexFormat
    {
        private Vector3 m_Position;
        private Color m_Color;
        private Vector2 m_UV;
        private Vector3 m_Normal;

        public static int SizeInBytes = (sizeof(float) * 8 + sizeof(uint));

        //Format in which this type of vertex will be passed to the graphics hardware
        public static VertexElement[] VertexElements =
        {
            new VertexElement(0, 0, VertexElementFormat.Vector3,
                VertexElementMethod.Default, VertexElementUsage.Position, 0),
            new VertexElement(0, sizeof(float)* 3, VertexElementFormat.Color,
                VertexElementMethod.Default, VertexElementUsage.Color, 0),
            new VertexElement(0, sizeof(float) * 3 + sizeof(uint),
                VertexElementFormat.Vector2, VertexElementMethod.Default,
                VertexElementUsage.TextureCoordinate, 0),
            new VertexElement(0, sizeof(float) * 2 + sizeof(uint),
                VertexElementFormat.Vector3, VertexElementMethod.Default,
                VertexElementUsage.Normal, 0),

        };

        //m_Position accessor and modifier
        public Vector3 Position
        {
            get
            {
                return m_Position;
            }
            set
            {
                m_Position = value;
            }
        }

        //m_Color accessor and modifier
        public Color Color
        {
            get
            {
                return m_Color;
            }
            set
            {
                m_Color = value;
            }
        }

        public float U
        {
            get { return m_UV.X; }
            set { m_UV.X = value; }
        }


        public float V
        {
            get { return m_UV.Y; }
            set { m_UV.Y = value; }
        }

        public Vector3 Normal
        {
            get { return m_Normal; }
            set { m_Normal = value; }
        }
    };

    public class XNAGraphicsDevice
    {
        private GraphicsDeviceManager m_Manager;
        private VertexDeclaration m_VertexDec;
        public ContentManager resources;
        public static XNAGraphicsDevice m_TheAPI = new XNAGraphicsDevice();
        RGameView serv;
        ContentManager content;
        private Viewport m_TempViewport;

        public UWB_XNAEffect m_LightingEffect;
        private UWB_XNALightManager m_LightManager;

        private XNAGraphicsDevice()
        {
        }

        public UWB_XNAEffect LightingEffect
        {
            get
            {
                return m_LightingEffect;
            }
        }

        public UWB_XNALightManager LightManager
        {
            get
            {
                return m_LightManager;
            }
        }

        public void SetService(RGameView SP)
        {
            serv = SP;
            resources = new ContentManager(serv);
            resources.RootDirectory = "Content";
        }

        public void CreateGraphicsContext(Game game, int deviceWidth, int deviceHeight)
        {
            m_Manager = new GraphicsDeviceManager(game);
            m_Manager.PreferredBackBufferWidth = deviceWidth;
            m_Manager.PreferredBackBufferHeight = deviceHeight;

            content = game.Content; //content = new ContentManager(Services);
        }

        public void Initialize()
        {
            m_LightingEffect = new UWB_XNAEffect(m_Manager.GraphicsDevice,
                content.Load<Effect>("Content/Effects/UWB_LightingEffect"));

            m_LightManager = new UWB_XNALightManager(m_LightingEffect);

            m_VertexDec = new VertexDeclaration(m_Manager.GraphicsDevice,
                DeviceVertexFormat.VertexElements);
            m_Manager.GraphicsDevice.VertexDeclaration = m_VertexDec;

            m_Manager.GraphicsDevice.RenderState.CullMode = CullMode.None;
            // m_Manager.GraphicsDevice.RenderState.MultiSampleAntiAlias = true;
            m_Manager.GraphicsDevice.RenderState.DepthBufferEnable = true;
            m_Manager.GraphicsDevice.RenderState.DepthBufferWriteEnable = true;
            // m_Manager.GraphicsDevice.RenderState.StencilEnable = true;

            m_LightingEffect.World = Matrix.Identity;
            m_LightingEffect.View = Matrix.Identity;
            m_LightingEffect.Projection = Matrix.Identity;
        }

        public void Clear(Color clearColor)
        {
            m_Manager.GraphicsDevice.Clear(clearColor);
        }

        public void Clear()
        {
            Clear(new Color(new Vector4(.8f, .8f, .8f, 1f)));
        }

        public void BeginDraw(Viewport viewport, Color backgroundColor)
        {
            m_TempViewport = m_Manager.GraphicsDevice.Viewport;

            m_Manager.GraphicsDevice.Viewport = viewport;
            m_Manager.GraphicsDevice.Clear(ClearOptions.Target | ClearOptions.DepthBuffer, backgroundColor, 1.0f, 0);
            m_Manager.GraphicsDevice.RenderState.DepthBufferEnable = true;
            m_Manager.GraphicsDevice.RenderState.DepthBufferWriteEnable = true;
        }

        public void EndDraw()
        {
            m_Manager.GraphicsDevice.Viewport = m_TempViewport;
        }

        public Matrix WorldMatrix
        {
            set
            {
                m_LightingEffect.World = value;
            }
        }

        public Matrix ViewMatrix
        {
            set
            {
                m_LightingEffect.View = value;
            }
        }

        public Matrix ProjectionMatrix
        {
            set
            {
                m_LightingEffect.Projection = value;
            }
        }

        public GraphicsDevice GraphicsDevice
        {
            get { return m_Manager.GraphicsDevice; }
        }

        public Viewport Viewport
        {
            set { m_Manager.GraphicsDevice.Viewport = value; }
            get { return m_Manager.GraphicsDevice.Viewport; }
        }

        public int DeviceHeight
        {
            get { return m_Manager.PreferredBackBufferHeight; }
        }

        public int DeviceWidth
        {
            get { return m_Manager.PreferredBackBufferWidth; }
        }

        public int DeviceTop
        {
            get { return 0; }
        }

        public int DeviceLeft
        {
            get { return 0; }
        }

        public void ShutDown()
        {
            //release all of the textures/meshes
            resources.Dispose();
        }

        /*
        public bool ActivateTexture(string texName)
        {
            Texture2D t = RetrieveTexture(texName);

            if (t != null) 
            {
                m_BasicEffect.Texture = t;
                m_BasicEffect.TextureEnabled = true;
                m_BasicEffect.VertexColorEnabled = false;
            }
            else
            {
                //texture load failed - disable texturing
                m_BasicEffect.Texture = null;
                m_BasicEffect.TextureEnabled = false;
                m_BasicEffect.VertexColorEnabled = true;
            }
            
            m_BasicEffect.CommitChanges();
            return true;
        }
        

        public void DeactivateTexture()
        {
            m_BasicEffect.TextureEnabled = false;
            m_BasicEffect.Texture = null;
            m_BasicEffect.VertexColorEnabled = true;
            m_BasicEffect.CommitChanges();
        }
        */

        public Texture2D RetrieveTexture(string texName)
        {
            Texture2D t = null;

            if (texName != null && texName != "")
            {
                try
                {
                    t = resources.Load<Texture2D>(texName);

                }
                catch (Exception e)
                {
                    //texture load failed
                    UWB_XNALib.UWB_Utility.echoToStatusArea((e.Message));
                }
            }
            return t;
        }

        public VertexDeclaration VertexDec
        {
            get
            {
                return m_VertexDec;
            }
            set
            {
                m_VertexDec = value;
            }
        }
    }
}
